# -*- coding: utf-8 -*-
"""
Created on Thu Sep 30 16:31:27 2021

@author: nkdi
"""

def load_packaged_model(model_file):
    import os
    import pickle
    # this_dir = os.getcwd()
    name,extension = os.path.splitext(model_file)
    if not extension:
        extension = '.sav'
    model_file = name + extension
    try:
        this_dir, this_filename = os.path.split(__file__)  # Get path of .pkl file
    except:
        this_dir = os.getcwd()
    data_path = os.path.join(this_dir, model_file)
    model = pickle.load(open(data_path, 'rb'))
    return model


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 24 12:47:19 2018

@author: nkdi
"""
import numpy as np
#from w2l import activation_functions
#from surrogates import math_utils
from hipersim.surrogates import math_utils

class ann:
    def __init__(self,**kwargs):
        #print(kwargs)
        if 'layersizes' in kwargs:
            self.layersizes = kwargs['layersizes']
        else:
            self.layersizes = [1, 10, 1]
            
        self.nlayers = len(self.layersizes) - 1 # Input layer is not counted
        
        if 'neurontypes' in kwargs:
            self.neurontypes = kwargs['neurontypes']
        else:
            self.neurontypes = ['Linear']*self.nlayers
            
        if 'activation' in kwargs:
            self.activation = kwargs['activation']
        else:
            self.activation = ['Relu']*(self.nlayers-1) + ['Linear']
            
        if 'testratios' in kwargs:
            self.testratios = kwargs['testratios']
        else:
            self.testratios = [0.7, 0.2, 0.1]
            
        if 'output_style' in kwargs:
            self.output_style = kwargs['output_style']
        else:
            self.output_style = 'None'
            
        if 'params' in kwargs:
            self.params = kwargs['params']
        else:
            self.params = {'weightinit': [0.01]*self.nlayers, 
                           'regularization': 0.1, 
                           'learningrate': 0.1, 
                           'costfunction': 'Lsq', 
                           'minibatchsize': 0, 
                           'maxiter': 1500, 
                           'RMSprop': 0.999, 
                           'Momentum': 0.9, 
                           'Xscaling': True, 
                           'Yscaling': True}
        # Parse parameters that may have not been given in the input
        if not 'weightinit' in self.params:
            self.params['weightinit'] = [0.01]*self.nlayers
        if not 'regularization' in self.params:
            self.params['regularization'] = 0.1
        if not 'learningrate' in self.params:           
            self.params['learningrate'] = 0.1
        if not 'costfunction' in self.params:
            self.params['costfunction'] = 'Lsq'
        if not 'minibatchsize' in self.params:
            self.params['minibatchsize'] = 0
        if not 'maxiter' in self.params:
            self.params['maxiter'] = 1500 # Same as nepochs
        if not 'RMSprop' in self.params:
            self.params['RMSprop'] = 0.999
        if not 'Momentum' in self.params:
            self.params['Momentum'] = 0.9
        if not 'nepochs' in self.params:
            self.params['nepochs'] = self.params['maxiter']
        if not 'Xscaling' in self.params:
            self.params['Xscaling'] = True
        if not 'Yscaling' in self.params:
            self.params['Yscaling'] = True


    def initialize_weights(self):
        W = []
        b = []
        #np.random.seed(1) # Used only in case we need to maintan consisteny with Matlab
        for i in range(self.nlayers):
            if self.neurontypes[i] == 'Linear':
                #Wi = np.random.randn(self.layersizes[i],self.layersizes[i+1])*self.params['weightinit'][i] # Regular way of initializing weights
                Wi = math_utils.NormalDist(2,np.random.rand(self.layersizes[i],self.layersizes[i+1]),0,1)*self.params['weightinit'][i] # For consistency with Matlab
                bi = np.zeros([1,self.layersizes[i+1]])
            elif self.neurontypes[i] == 'LinearPretrained':
                Wi = self.weights['W'][i]
                bi = self.weights['bias'][i]
            elif self.neurontypes[i] == 'LinearNonTrainable':
                Wi = self.weights['W'][i]
                bi = self.weights['bias'][i]                
            W.append(Wi)
            b.append(bi)
            
        if hasattr(self, 'weights'):
            self.weights['W'] = W    
            self.weights['bias'] = b
        else:
            self.weights = {'W':W, 'bias':b}
                        
        return W,b
    
    def forward_propagation(self,X):
        A = []
        Z = []
        A.append(X)
        Ai = X
        for i in range(self.nlayers):
            if self.neurontypes[i]=='Linear':
                Zi = np.dot(Ai,self.weights['W'][i]) + self.weights['bias'][i]
            elif self.neurontypes[i]=='LinearPretrained':
                Zi = np.dot(Ai,self.weights['W'][i]) + self.weights['bias'][i]
            elif self.neurontypes[i]=='LinearNonTrainable':
                Zi = np.dot(Ai,self.weights['W'][i]) + self.weights['bias'][i]
            else:
                raise ValueError('Unknown or unspecified neuron type')

            if self.activation[i] == 'Relu':
                Ai = math_utils.relu(Zi)
            elif self.activation[i] == 'Linear':
                Ai = Zi
            elif self.activation[i] == 'Tanh':
                Ai = math_utils.hyperbolic_tangent(Zi)
            elif self.activation[i] == 'Softplus':
                Ai = math_utils.softplus(Zi)
            else:
                raise ValueError('Unknown or unspecified activation function')
            
            A.append(Ai)
            Z.append(Zi)
            
        Ypredict = Ai
        ANNcache = (A,Z)
        return Ypredict, ANNcache
    
    def backward_propagation(self,X,Y,Ypredict,ANNcache):
        if self.params['costfunction'] == 'Log':
            dJdAprev = -Y/Ypredict + (1 - Y)/(1 - Ypredict)
        elif self.params['costfunction'] == 'Lsq':
            dJdAprev = (Ypredict -Y)
        else:
            raise ValueError('Unknown cost function type')
        
        dJdA = [None]*self.nlayers
        dAdZ = [None]*self.nlayers
        dJdW = [None]*self.nlayers
        dJdb = [None]*self.nlayers 
        dJdZ = [None]*self.nlayers
        A,Z = ANNcache
        
        for i in range(self.nlayers-1, -1,-1):
            dJdA[i] = dJdAprev
            if self.activation[i] == 'Linear':
                dAdZ[i] = np.ones(Z[i].shape)
            elif self.activation[i] == 'Relu':
                dAdZ[i] = np.ones(Z[i].shape)*(Z[i] > 0)
            elif self.activation[i] == 'Softplus':
                dAdZ[i] = math_utils.logistic(Z[i])
            elif self.activation[i] == 'Tanh':
                dAdZ[i] = 1 - math_utils.hyperbolic_tangent(Z[i])**2
            elif self.activation[i] == 'Logistic':
                dAdZ[i] = math_utils.logistic(Z[i])*(1 - math_utils.logistic(Z[i]))
            else:
                raise ValueError('Unknown activation function')
            
            dJdZ[i] = dAdZ[i]*dJdA[i]
            mlength = X.shape[0]
            if self.neurontypes[i] == 'Linear':
                dJdW[i] = (1/mlength)*np.dot(A[i].T,dJdZ[i])
                dJdb[i] = (1/mlength)*np.sum(dJdZ[i],0,keepdims = True)
            elif self.neurontypes[i] == 'LinearPretrained':
                dJdW[i] = (1/mlength)*np.dot(A[i].T,dJdZ[i])
                dJdb[i] = (1/mlength)*np.sum(dJdZ[i],0,keepdims = True)
            elif self.neurontypes[i] == 'LinearNonTrainable':
                dJdW[i] = np.zeros(Z[i].shape)
                dJdb[i] = np.zeros((1,Z[i].shape[1]))
            else:
                raise ValueError('Unknown neuron type')

            dJdAprev = np.dot(dJdZ[i],self.weights['W'][i].T)            
        
        return dJdW, dJdb

    def backward_propagation_dy(self,X):
        if self.params['Xscaling']:
            X = (X - self.Xbias)/self.Xscale # Lots of broadcasting taking place here
            
        Y, ANNcache = self.forward_propagation(X)
        if self.params['Yscaling']:
            Y = Y*self.Yscale + self.Ybias
        
        A,Z = ANNcache
        noutputs = Y.shape[1]
                
        dY = [None]*noutputs
        
        for iY in range(noutputs):
        
            for i in range(self.nlayers-1, -1,-1):
                
                if i == self.nlayers - 1:
                    Zi = Z[i][:,iY]
                    Zi = Zi.reshape(Zi.shape[0],1)
                    Wi = self.weights['W'][i][:,iY]
                    Wi = Wi.reshape(Wi.shape[0],1)
                else:
                    Zi = Z[i]
                    Wi = self.weights['W'][i]
                    
                if self.activation[i] == 'Linear':
                    dAdZ = np.ones(Zi.shape)
                elif self.activation[i] == 'Relu':
                    dAdZ = np.ones(Zi.shape)*(Zi > 0)
                elif self.activation[i] == 'Softplus':
                    dAdZ = math_utils.logistic(Zi)
                elif self.activation[i] == 'Tanh':
                    dAdZ = 1 - math_utils.hyperbolic_tangent(Zi)**2
                elif self.activation[i] == 'Logistic':
                    dAdZ = math_utils.logistic(Zi)*(1 - math_utils.logistic(Zi))
                else:
                    raise ValueError('Unknown activation function')
                
                if i == self.nlayers-1:
                    dYi = np.dot(dAdZ,Wi.T)
                else:
                    dYi = np.dot((dAdZ*dYi),Wi.T)

            if self.params['Yscaling']:
                dYi = dYi*self.Yscale[0,iY]
            if self.params['Xscaling']:
                dYi = dYi/self.Xscale
            
            dY[iY] = dYi
                        
        return Y, dY
    
    def check_gradients(self,X):
        if not hasattr(self,'weights'):
            self.initialize_weights()
            
            
    def check_gradients_dxdy(self,X):
        Yout,dYout = self.backward_propagation_dy(X)
    
    def compute_cost(self,X,Y):
        #Computes the cost using log-functions
        mtrain = X.shape[0]
        noutputs = Y.shape[1]
        Ypredict,ANNcache = self.forward_propagation(X)
        if self.params['costfunction'] == 'Log':
            Loss = - (np.multiply(Y,np.log(Ypredict)) + np.multiply((1 - Y), np.log(1 - Ypredict)))
        elif self.params['costfunction'] == 'Lsq':
            Loss = 0.5*np.multiply((Ypredict - Y),(Ypredict - Y))
        else:
            print('Unknown loss function type!')

        if self.params["regularization"] == 0:
            J = (1/mtrain)*np.sum(Loss)*(1/noutputs)
        else:
            v = cell_to_vector(self.weights['W'], self.weights['bias'])
            J = (1/mtrain)*np.sum(Loss)*(1/noutputs) + (self.params['regularization']/(2*mtrain))*(np.dot(v.T,v))

        return J, Ypredict, ANNcache

    
    def train(self,X,Y):
        """
        Uses input-output data pair (X,Y) to train the weights of a neural-net object
        """        
        # Parse hyperparameters
        beta2 = self.params['RMSprop']
        if beta2 == 0:
            epsroundoff = 0
        else:
            epsroundoff = 1e-8
        beta1 = self.params['Momentum']                    
        minibatchsize = self.params['minibatchsize']
        nepochs = self.params['nepochs']
                    
        self.Xbias = np.mean(X,axis=0, keepdims = True) # Keepdims = True needed to allow broadcasting
        self.Ybias = np.mean(Y,axis=0, keepdims = True)
        self.Xscale = np.std(X,axis=0, ddof = 1, keepdims = True)
        self.Yscale = np.std(Y,axis=0, ddof = 1, keepdims = True)
        
        if self.params['Xscaling']:
            X = (X - self.Xbias)/self.Xscale
        if self.params['Yscaling']:
            Y = (Y - self.Ybias)/self.Yscale

        # Split data into training, validation and test sets            
        Ntotal = X.shape[0]
        Ntrain = int(np.floor(Ntotal*self.testratios[0]))
        Ndev = int(np.floor(Ntotal*self.testratios[1]))
        Xtrain = X[:Ntrain,:]
        Ytrain = Y[:Ntrain,:]
        Xdev = X[Ntrain:Ntrain+Ndev,:]
        Ydev = Y[Ntrain:Ntrain+Ndev,:]
        Xtest = X[Ntrain+Ndev:,:]
        Ytest = Y[Ntrain+Ndev:,:]

        # Determine iteration scheme
        if minibatchsize == 0: # Minibatch functionality off
            minibatchsize = Ntrain # Each iteration is on the entire train set            
            Nbatches = 1
            Niter = self.params['nepochs']
        else:
            Nbatches = int(np.ceil(Ntrain/minibatchsize))
            Niter = nepochs*Nbatches
                        
        # Initialize variables and weights
        noutputs = Y.shape[1]
        Jhist = np.zeros([Niter,1])
        Jgradnorm = np.zeros([Niter,1])
        Jdevhist = np.zeros([Niter,1])
        Rsqhist = np.zeros([Niter,noutputs])
        Rsqdevhist = np.zeros([Niter,noutputs])
        
        self.initialize_weights()
        ntotal = 0
        for i in range(self.nlayers):
            ntotal+= self.weights['W'][i].shape[0]*self.weights['W'][i].shape[1] + self.weights['bias'][i].shape[0]*self.weights['bias'][i].shape[1] 
        
        # Initialize interactive plot/inline-output functionalities
        if self.output_style == 'Verbose':
            print('Verbose output')
        elif self.output_style == 'Plot':
            import tkinter
            import matplotlib            
            import matplotlib.pyplot as plt
            plt.ion()
            matplotlib.use('TkAgg')
            fig, axs = plt.subplots(1,2,figsize = (8,4))
            plt00 = axs[0].plot(0,0,'-k')[0]
            plt01 = axs[0].plot(0,0,'-r')[0]
            axs[0].set_xlabel('Number of iterations')
            axs[0].set_ylabel('Cost function value J')
            plt10 = axs[1].plot(0,0,'-k')[0]
            plt11 = axs[1].plot(0,0,'-r')[0]
            plt12 = axs[1].plot(0,0,'--k')[0]
            plt13 = axs[1].plot(0,0,'--r')[0]
            
            axs[1].set_xlabel('Number of iterations')
            axs[1].set_ylabel('R-squared')
            axs[1].set_ylim([0,1])
            fig.tight_layout()
            fig.show()
            fig.canvas.draw()
            
        elif self.output_style == 'None': 
            print('Training neural net model:')
            
            
        # Begin iterations
        go = 1
        i = 0
        ibatch = 0
        iepoch = 0
        Vd = np.zeros([ntotal,1])
        Sd = np.zeros([ntotal,1])
        while go == 1:
            # Look at the dev(validation) set and compute cost function and R-square
            Jdev,Ypredictdev, _ = self.compute_cost(Xdev,Ydev)
            Jdevhist[i] = Jdev
            Rsqdevhist[i,:] = math_utils.pairwise_correlation(Ypredictdev,Ydev)**2
            
            # Pick a single minibatch of training data samples
            if ibatch == Nbatches:
                Xtraini = Xtrain[ibatch*minibatchsize:,:]
                Ytraini = Ytrain[ibatch*minibatchsize:,:]
            else:
                Xtraini = Xtrain[ibatch*minibatchsize:(ibatch+1)*minibatchsize,:]
                Ytraini = Ytrain[ibatch*minibatchsize:(ibatch+1)*minibatchsize,:]
                
            # Compute cost function, cost function gradients, and R-square on one minibatch
            J,Ypredict,ANNcache = self.compute_cost(Xtraini,Ytraini) # Cost function
            [dJdW,dJdb] = self.backward_propagation(Xtraini,Ytraini,Ypredict,ANNcache) # Cost function gradients
            dvecti = cell_to_vector(dJdW,dJdb)
            Jhist[i] = J
            Jgradnorm[i] = (1/minibatchsize)*np.sqrt(np.dot(dvecti.T,dvecti)) # Gradient norm for monitoring purposes
            Rsqhist[i,:] = math_utils.pairwise_correlation(Ypredict,Ytraini)**2 # R-square
            
            # Update weights based on gradient descent
            if (beta1 != 0) & (beta2 != 0): # Adam algorithm
                Vd = beta1*Vd + (1 - beta1)*dvecti
                Sd = beta2*Sd + (1 - beta2)*(dvecti**2)
                Vcorr = Vd/(1 - beta1**(i+1))
                Scorr = Sd/(1 - beta2**(i+1))
                vupd = cell_to_vector(self.weights['W'], self.weights['bias']) - self.params['learningrate']*(Vcorr/(np.sqrt(Scorr) + epsroundoff)  + cell_to_vector(self.weights['W'], self.weights['bias'])*self.params['regularization']/Xtraini.shape[0])
            elif beta1 > 0 & beta2 == 0: # Only momentum, no RMSprop, optional regularization
                Vd = beta1*Vd + (1 - beta1)*dvecti
                Vcorr = Vd/(1 - beta1**(i+1))
                vupd = cell_to_vector(self.weights['W'], self.weights['bias']) - self.params['learningrate']*(Vcorr + cell_to_vector(self.weights['W'], self.weights['bias'])*self.params['regularization']/Xtraini.shape[0])
            else:
                # Simple gradient descent with optional regularization
                vupd = cell_to_vector(self.weights['W'], self.weights['bias']) - self.params['learningrate']*(cell_to_vector(dJdW,dJdb) + cell_to_vector(self.weights['W'], self.weights['bias'])*self.params['regularization']/Xtraini.shape[0])

            self.weights['W'], self.weights['bias'] = vector_to_cell(vupd,self.weights['W'], self.weights['bias'])
            
            # Print/plot iteration outputs
            if self.output_style == 'Verbose':
                print('Epoch #' + str(iepoch) + ', batch #' + str(ibatch) + ', cost function value = ' + str(J[0]))
            elif (self.output_style == 'Plot') & (i >=1):
                
                # Update plot with results from current iteration
                xplot = np.arange(i)
                plt00.set_data(xplot,Jhist[:i])
                plt01.set_data(xplot,Jdevhist[:i])
                axs[0].relim()
                axs[0].autoscale_view(True,True,True) #Autoscale                
                plt10.set_data(xplot,np.min(Rsqhist[:i,:],1))
                plt11.set_data(xplot,np.min(Rsqdevhist[:i,:],1))
                plt12.set_data(xplot,np.max(Rsqhist[:i,:],1))
                plt13.set_data(xplot,np.max(Rsqdevhist[:i,:],1))
                
                axs[1].relim()
                axs[1].autoscale_view(True,True,True) #Autoscale                
                
                plt.pause(0.001)
                fig.canvas.draw()
                plt.show()

            
            i+=1
            ibatch+=1
            if i >= Niter:
                go = 0
            if ibatch >= Nbatches:
                if iepoch >= nepochs:
                    go = 0
                else:
                    ibatch=0
                    iepoch+=1
                    
        Jout,Yout, _ = self.compute_cost(Xtest,Ytest)    
        Rsqout = math_utils.pairwise_correlation(Yout,Ytest)**2
        
        if self.params['Yscaling']:
            Testbias = (np.mean(Yout,0) - np.mean(Ytest,0))*self.Yscale
        else:
            Testbias = np.mean(Yout,0) - np.mean(Ytest,0)
        
        self.metrics = {'RsqTest':Rsqout, 'RsqDev':Rsqdevhist[i-1,:],'RsqTrain':Rsqhist[i-1,:],'Testbias':Testbias}
        
        self.neurontypes = ['LinearPretrained' if ni == 'Linear' else ni for ni in self.neurontypes] # Start from previous trained state in case a new training command is issued
                    
        Outdata = {'Jhist':Jhist, 'Yout':Yout, 'Jout':Jout}
        
                
        return Outdata
    
    def predict(self,X):
        """
        Run a model prediction (forward propagation + scaling)
        """        
        if self.params['Xscaling']:
            X = (X - self.Xbias)/self.Xscale # Lots of broadcasting taking place here
        
        Y, _ = self.forward_propagation(X)
        if self.params['Yscaling']:
            Y = Y*self.Yscale + self.Ybias # Lots of broadcasting also taking place here
        
        return Y
        
def cell_to_vector(W,b):
    """
    Reshapes an array from dictionary-type structure to vector
    """
    nlayers = len(W)
    ntotal = 0
    for i in range (nlayers):
        ntotal+= W[i].shape[0]*W[i].shape[1] + b[i].shape[0]*b[i].shape[1]        
    v = np.zeros([ntotal,1])
    vcount = 0
    for i in range(nlayers):
        vinc = W[i].shape[0]*W[i].shape[1]
        v[vcount:vcount+vinc] = np.reshape(W[i],(vinc,1))
        vcount+= vinc
    for i in range(nlayers):
        vinc = b[i].shape[0]*b[i].shape[1]
        v[vcount:vcount+vinc] = np.reshape(b[i],(vinc,1))
        vcount+= vinc
        
    return v
        
def vector_to_cell(v,W,b):
    """
    Reshapes an array from a 1-D vector to a dictionary-type structure
    """

    nlayers = len(W)
    vcount = 0
    
    for i in range(nlayers):
        vinc = W[i].shape[0]*W[i].shape[1]
        W[i] = np.reshape(v[vcount:vcount+vinc],(W[i].shape[0],W[i].shape[1]))
        vcount+= vinc
    
    for i in range(nlayers):
        vinc = b[i].shape[0]*b[i].shape[1]
        b[i] = np.reshape(v[vcount:vcount+vinc],(b[i].shape[0],b[i].shape[1]))
        vcount+= vinc
    
    return W, b   





# -*- coding: utf-8 -*-
"""
Created on Wed Nov 10 14:38:33 2021

@author: nkdi
"""

def output_field(u0,v0,w0,params,TurbOptions = None):
    
    import numpy as np
    import struct    
    
    if TurbOptions is None:
        TurbOptions = {'FileFormat':0}               
    
    # APPLY MEAN AND VARIANCE CORRECTION
    if 'Umean' in TurbOptions:
        MFFWS = TurbOptions['Umean']
    else:
        MFFWS = 0
        
    if 'TI_u' in TurbOptions:        
        TI_U = TurbOptions['TI_u']
        TI_V = TurbOptions['TI_v']
        TI_W = TurbOptions['TI_w']
                
        u = (u0 - np.mean(u0))*(TI_U*MFFWS/np.std(u0))
        v = (v0 - np.mean(v0))*(TI_V*MFFWS/np.std(v0))
        w = (w0 - np.mean(w0))*(TI_W*MFFWS/np.std(w0))    
    else:
        # Just zero-mean
        u = (u0 - np.mean(u0))
        v = (v0 - np.mean(v0))
        w = (w0 - np.mean(w0))        
    
    
    # BOX DIMENSIONS AND SPACING
    if 'Nx' in params:
        Nx = params['Nx']
    else:
        Nx = u.shape[0]
    if 'Ny' in params:
        Ny = params['Ny']
    else:
        Ny = u.shape[1]
    if 'Nz' in params:
        Nz = params['Nz']
    else:
        Nz = u.shape[2]             

    dx = params['dx']
    dy = params['dy']
    dz = params['dz']           
        
    # APPLY SHEAR CORRECTION IF NECESSARY                    
    
    if 'ShearLaw' in TurbOptions:
        ShearLaw = TurbOptions['ShearLaw']
    else:
        ShearLaw = 'pwr'
        
    if 'zHub' in TurbOptions:
        zHub = TurbOptions['zHub']
    else:
        zHub = dz*(Nz-1)/2
            
    zGoffset = 0.0 # Can in principle be an input
    z1   = zHub - zGoffset - dz*(Nz-1)/2 #  this is the bottom of the grid
    zbox = np.arange(Nz)*dz + z1
    
    if ShearLaw == 'pwr':
        alpha = TurbOptions['alpha']
        zOffset = zHub
        z0 = np.exp(( - np.log(zOffset)*(1/zOffset)**alpha)/(1 - (1/zOffset)**alpha)) # Roughness length [m]
        ShearMultiplier = np.tile((zbox/zHub)**alpha, (Nx,Ny,1))
    elif ShearLaw == 'log':
        z0 = TurbOptions['z0']
        zOffset = zHub
        ShearMultiplier = np.tile(np.log(zbox/z0)/np.log(zOffset/z0), (Nx,Ny,1))
        
    u = u + MFFWS*ShearMultiplier
    
    # ROTATE FIELDS IF NECESSARY
    if 'Yaw' in TurbOptions:
        YawAngle = TurbOptions['Yaw']
    else:
        YawAngle = 0
        
    if 'Pitch' in TurbOptions:
        PitchAngle = TurbOptions['Pitch']
    else:
        PitchAngle = 0    
        
    if 'Roll' in TurbOptions:
        PitchAngle = TurbOptions['Roll']
    else:
        RollAngle = 0      
    
    if abs(YawAngle) + abs(PitchAngle) + abs(RollAngle) > 0:
        pi = np.pi
        c0 = np.cos(YawAngle*pi/180)
        s0 = np.sin(YawAngle*pi/180)
        c1 = np.cos(PitchAngle*pi/180)
        s1 = np.sin(PitchAngle*pi/180)
        c2 = np.cos(RollAngle*pi/180)
        s2 = np.sin(RollAngle*pi/180)    
        
        RotationMatrix = np.zeros((3,3))
        RotationMatrix[0,0] = c0*c1
        RotationMatrix[0,1] = c0*s1*s2 - s0*c2
        RotationMatrix[0,2] = c0*s1*c2 + s0*s2
        RotationMatrix[1,0] = s0*c1
        RotationMatrix[1,1] = s0*s1*s2 + c0*c2
        RotationMatrix[1,2] = s0*s1*c2 - c0*s2
        RotationMatrix[2,0] = -s1
        RotationMatrix[2,1] = c1*s2
        RotationMatrix[2,2] = c1*c2
        
        urot = RotationMatrix[0,0]*u + RotationMatrix[0,1]*v + RotationMatrix[0,2]*w
        vrot = RotationMatrix[1,0]*u + RotationMatrix[1,1]*v + RotationMatrix[1,2]*w
        wrot = RotationMatrix[2,0]*u + RotationMatrix[2,1]*v + RotationMatrix[2,2]*w
        
        u = urot
        v = vrot
        w = wrot             
            
    if 'FileFormat' not in TurbOptions:
        FileFormat = 0
    else:
        FileFormat = TurbOptions['FileFormat']      
        
    if 'BaseName' in params:
        BaseName = params['BaseName']
    else:
        BaseName = 'turb'
        
    if 'SeedNo' in params:
        SeedNo = params['SeedNo']
    else:
        SeedNo = 1       
        
    if 'SaveToFile' in params:
        SaveToFile = params['SaveToFile']
    else:
        SaveToFile = 0
        
    
    if (SaveToFile == 1) | (SaveToFile == True):
        
        if (FileFormat == 0) | (FileFormat == 'Hawc2') | (FileFormat == 'Mann'):
            u.reshape(Nx*Ny*Nz).astype('single').tofile(BaseName + '_' + str(SeedNo) + '_u' + '.bin',sep = '')
            v.reshape(Nx*Ny*Nz).astype('single').tofile(BaseName + '_' + str(SeedNo) + '_v' + '.bin',sep = '')
            w.reshape(Nx*Ny*Nz).astype('single').tofile(BaseName + '_' + str(SeedNo) + '_w' + '.bin',sep = '')
        elif (FileFormat == 'Bladed') | (FileFormat == 'wnd') | (FileFormat == 'TurbSim'):                
        
            Scale    = 0.00001*MFFWS*np.asarray([TI_U*100, TI_V*100, TI_W*100])
            Offset   = np.array([MFFWS,0,0])
            
            
            uOut = np.asarray((u - Offset[0])/Scale[0], dtype = 'int16')
            vOut = np.asarray((v - Offset[1])/Scale[1], dtype = 'int16')
            wOut = np.asarray((w - Offset[2])/Scale[2], dtype = 'int16')
            
            OutFileName = str(BaseName + '_' + str(SeedNo) + '.wnd')
            wnd_file = open(OutFileName,'wb')
            
            # HEADER OF THE .WND FILE
            wnd_file.write(struct.pack('<h',-99))             # ID - must be -99
            wnd_file.write(struct.pack('<h',4))               # ID2 - must be 4
            wnd_file.write(struct.pack('<i', 3))              # number of components (should be 3)
            wnd_file.write(struct.pack('<f',0.0))             # latitude (deg)
            wnd_file.write(struct.pack('<f',z0))              # Roughness length (m)
            wnd_file.write(struct.pack('<f',zOffset))         # Reference height (m) = Z(1) + GridHeight / 2.0
            wnd_file.write(struct.pack('<f',TI_U*100))        # Turbulence Intensity of u component (%)
            wnd_file.write(struct.pack('<f',TI_V*100))        # Turbulence Intensity of v component (%)
            wnd_file.write(struct.pack('<f',TI_W*100))        # Turbulence Intensity of w component (%)
            wnd_file.write(struct.pack('<f',dz))              # delta z in m 
            wnd_file.write(struct.pack('<f',dy))              # delta y in m 
            wnd_file.write(struct.pack('<f',dx))              # delta x in m 
            wnd_file.write(struct.pack('<i',int(Nx/2)))       # half the number of time steps (points in longitudinal direction)
            wnd_file.write(struct.pack('<f',MFFWS))           # mean full-field wind speed
            wnd_file.write(struct.pack('<f',0.0))             # zLu - unused variable
            wnd_file.write(struct.pack('<f',0.0))             # yLu - unused variable
            wnd_file.write(struct.pack('<f',0.0))             # xLu - unused variable
            wnd_file.write(struct.pack('<i',0))               # _ - unused variable
            wnd_file.write(struct.pack('<i',0))               # RandSeed1 - unused variable
            wnd_file.write(struct.pack('<i',Nz))              # number of points in vertical direction
            wnd_file.write(struct.pack('<i',Ny))              # number of points in horizontal direction
            wnd_file.write(struct.pack('<i',0))               # Unused variable - for BLADED
            wnd_file.write(struct.pack('<i',0))               # Unused variable - for BLADED
            wnd_file.write(struct.pack('<i',0))               # Unused variable - for BLADED
            wnd_file.write(struct.pack('<i',0))               # Unused variable - for BLADED
            wnd_file.write(struct.pack('<i',0))               # Unused variable - for BLADED
            wnd_file.write(struct.pack('<i',0))               # Unused variable - for BLADED
            
            
            # WRITE WIND FIELD DATA OUT
            Clockwise = True # Should be loaded from the summary file in principle
            if Clockwise == True:
                y_ix = np.flip(np.arange(Ny))
            else:
                y_ix = np.arange(Ny)
            
            for it in range(Nx):
                for iz in range(Nz):
                    for iy in y_ix:
                        wnd_file.write(struct.pack('<h',uOut[it,iy,iz]))
                        wnd_file.write(struct.pack('<h',vOut[it,iy,iz]))
                        wnd_file.write(struct.pack('<h',wOut[it,iy,iz]))
            
            wnd_file.close()
    
    return u,v,w


# -*- coding: utf-8 -*-
"""
Created on Mon Sep 27 23:07:00 2021

@author: nkdi
"""
import numpy as np
def TrigonometricToAzimuth(Theta,AzimuthToTrigonometric = None,Rad = None):

    if AzimuthToTrigonometric is None:
        AzimuthToTrigonometric = 0

    if Rad is None:
        Rad = 0

    if Rad == 1:
        ThetaDiff = np.pi/2
        ThetaMax = 2*np.pi
        ThetaMin = 0
    else:
        ThetaDiff = 90
        ThetaMax = 360
        ThetaMin = 0

    if AzimuthToTrigonometric == 0:
        ThetaOut = -np.asarray(Theta) + ThetaDiff
    elif AzimuthToTrigonometric == 1:
        ThetaOut = ThetaDiff - np.asarray(Theta)
        
    ThetaOut = np.asarray(ThetaOut) # Avoid problems when evaluating scalars

    ThetaOut[ThetaOut < ThetaMin] = ThetaOut[ThetaOut < ThetaMin] + ThetaMax
    ThetaOut[ThetaOut >= ThetaMax] = ThetaOut[ThetaOut >= ThetaMax] - ThetaMax

    return ThetaOut

def FarmUpwindDisturbances(ThetaRange,TurbineIDs,TurbinePosLon,TurbinePosLat,OutputTurbineIDs=None,ThetaRelRange=None):

    if OutputTurbineIDs is None:
        OutputTurbineIDs = TurbineIDs
    if ThetaRelRange is None:
        ThetaRelRange = 16
    TurbineIDsNumeric = np.atleast_2d(np.arange(len(TurbineIDs))).T

    DX = np.dot(np.atleast_2d(TurbinePosLon).T,np.ones((1,len(TurbinePosLon)))) - np.dot(np.ones((len(TurbinePosLon),1)),np.atleast_2d(TurbinePosLon))
    DY = np.dot(np.atleast_2d(TurbinePosLat).T,np.ones((1,len(TurbinePosLat)))) - np.dot(np.ones((len(TurbinePosLat),1)),np.atleast_2d(TurbinePosLat))

    Theta = np.arctan2(DY,DX)
    Theta = Theta + (Theta < 0)*2*np.pi
    Theta = Theta - (Theta > 2*np.pi)*2*np.pi
    ThetaDeg = Theta*180/np.pi
    Rdist = np.sqrt(DX**2 + DY**2)

    NOutputTurbines = len(OutputTurbineIDs)
    ClosestDisturbanceID = np.zeros((len(ThetaRange),NOutputTurbines))
    ClosestDisturbanceDist = np.zeros((len(ThetaRange),NOutputTurbines))
    ClosestDisturbanceRelAngle = np.zeros((len(ThetaRange),NOutputTurbines))


    DisturbanceList = [[] for _ in range(len(ThetaRange))]
    DisturbanceDist = [[] for _ in range(len(ThetaRange))]
    DisturbanceAzimuth = [[] for _ in range(len(ThetaRange))]
    DisturbanceRelAngle = [[] for _ in range(len(ThetaRange))]
    ClosestDisturbanceID = [[] for _ in range(len(ThetaRange))]
    ClosestDisturbanceDist = np.zeros((len(ThetaRange),len(OutputTurbineIDs)))
    ClosestDisturbanceRelAngle = np.zeros((len(ThetaRange),len(OutputTurbineIDs)))

    for iT in range(len(ThetaRange)):
        Thetai = TrigonometricToAzimuth(ThetaRange[iT],1,0) # Azimuth to trigonometric in degrees    
        DisturbanceList[iT] = [[] for _ in range(len(OutputTurbineIDs))]
        DisturbanceDist[iT] = [[] for _ in range(len(OutputTurbineIDs))]
        DisturbanceAzimuth[iT] = [[] for _ in range(len(OutputTurbineIDs))]
        DisturbanceRelAngle[iT] = [[] for _ in range(len(OutputTurbineIDs))]
        ClosestDisturbanceID[iT] = [[] for _ in range(len(OutputTurbineIDs))]
        for iO in range(len(OutputTurbineIDs)):
            CurrentTurbine = np.argwhere(OutputTurbineIDs[iO]==TurbineIDs)[0]
            ThetaC = ThetaDeg[:,CurrentTurbine]
            RdistC = Rdist[:,CurrentTurbine]
            Rindex = RdistC>0
            ThetaDegW = ThetaC[Rindex == True]
            RdistW = RdistC[Rindex==True]
            TurbineIDsW = TurbineIDsNumeric[Rindex==True]    
            Tindex = ((ThetaDegW >= (Thetai - ThetaRelRange)) & (ThetaDegW < (Thetai + ThetaRelRange))) | ((ThetaDegW > (Thetai - ThetaRelRange + 360)) | (ThetaDegW < (Thetai + ThetaRelRange - 360)))
            if sum(Tindex==True) > 0:
                ThetaDegSector = ThetaDegW[Tindex]
                RdistSector = RdistW[Tindex]
                TurbineIDsSector = TurbineIDsW[Tindex]
                DistanceSortIndex = np.argsort(RdistSector)
                iDistancesSort = RdistSector[DistanceSortIndex]
                DisturbanceList[iT][iO] = TurbineIDs[TurbineIDsSector[DistanceSortIndex]]
                DisturbanceDist[iT][iO] = iDistancesSort
                DisturbanceAzimuth[iT][iO] = TrigonometricToAzimuth(ThetaDegSector[DistanceSortIndex])
                RelativeAnglei = ThetaRange[iT] - TrigonometricToAzimuth(ThetaDegSector[DistanceSortIndex])
                RelativeAnglei = RelativeAnglei - (RelativeAnglei > 180)*360
                RelativeAnglei = RelativeAnglei + (RelativeAnglei < -180)*360
                DisturbanceRelAngle[iT][iO] = RelativeAnglei                        
                ClosestDisturbanceID[iT][iO] = TurbineIDs[TurbineIDsSector[DistanceSortIndex[0]]]
                ClosestDisturbanceDist[iT,iO] = iDistancesSort[0]
                ClosestDisturbanceRelAngle[iT,iO] = RelativeAnglei[0]
    
    return DisturbanceList,DisturbanceDist,DisturbanceRelAngle,DisturbanceAzimuth,ClosestDisturbanceID,ClosestDisturbanceDist,ClosestDisturbanceRelAngle

def CircularMean_D(x):
    import numpy as np
    x = np.asarray(x[~np.isnan(np.asarray(x,dtype = 'float64'))],dtype = 'float64')
    mu = np.arctan2(np.mean(np.sin(x*np.pi/180)),np.mean(np.cos(x*np.pi/180)))*180/np.pi
    mu+=360*((mu<=-0.5)==True)
    return mu

def EncodeWindFarmGeometry(TurbinePosLon,TurbinePosLat,ThetaRange, method = None, inputs = None):
    from hipersim.surrogates.farm_utils import FarmUpwindDisturbances
    if method is None:
        method = 'RelativeLocation'


    Nturbines = len(TurbinePosLon)
    Nthetas = len(ThetaRange)
    TurbineIDNumeric = np.arange(1,Nturbines+1)        
    
    if method == 'RelativeLocation':
        if inputs is None:
            inputs = dict()
        if 'Nclosest' not in inputs:
            inputs['Nclosest'] = 10        
        Nclosest = inputs['Nclosest']
        DisturbanceList,DisturbanceDist,DisturbanceRelAngle,__,__,__,__ = FarmUpwindDisturbances(ThetaRange,TurbineIDNumeric,TurbinePosLon,TurbinePosLat,None,None)    
        X = np.zeros((Nturbines*Nthetas,Nclosest*2))
        for iTheta in range(Nthetas):
            for iTurb in range(Nturbines):
                nDisturb = len(DisturbanceList[iTheta][iTurb])
                nDisturb = np.min((nDisturb,Nclosest))
                X[iTheta*Nturbines + iTurb,:nDisturb] = DisturbanceDist[iTheta][iTurb][:nDisturb]
                X[iTheta*Nturbines + iTurb,Nclosest:(Nclosest + nDisturb)] = np.sin(np.deg2rad(DisturbanceRelAngle[iTheta][iTurb][:nDisturb]))*DisturbanceDist[iTheta][iTurb][:nDisturb]
    
    if method == 'Autoencoder':
        from hipersim.surrogates.os_utils import load_packaged_model
        if inputs is None:
            inputs = dict()
        if 'Modelfile' not in inputs:
            inputs['Modelfile'] = 'AutoencoderModel_v0.sav'            
        if 'Nclosest' not in inputs:
            inputs['Nclosest'] = 20            
        Nclosest = inputs['Nclosest']
        DisturbanceList,DisturbanceDist,DisturbanceRelAngle,__,__,__,__ = FarmUpwindDisturbances(ThetaRange,TurbineIDNumeric,TurbinePosLon,TurbinePosLat,None,None)    
        X = np.zeros((Nturbines*Nthetas,Nclosest*2))
        for iTheta in range(Nthetas):
            for iTurb in range(Nturbines):
                nDisturb = len(DisturbanceList[iTheta][iTurb])
                nDisturb = np.min((nDisturb,Nclosest))
                X[iTheta*Nturbines + iTurb,:nDisturb] = DisturbanceDist[iTheta][iTurb][:nDisturb]
                X[iTheta*Nturbines + iTurb,Nclosest:(Nclosest + nDisturb)] = np.sin(np.deg2rad(DisturbanceRelAngle[iTheta][iTurb][:nDisturb]))*DisturbanceDist[iTheta][iTurb][:nDisturb]
        
        ANN = load_packaged_model(inputs['Modelfile'])
        X = ANN.predict(X)
    
    return X


# -*- coding: utf-8 -*-
"""
Created on Sat Apr 24 09:57:06 2021

@author: nkdi
"""
import numpy as np
def manntensor(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel,ElementChoice = None):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    kL = np.asarray(k*L)
    k1square = k1**2
    k2square = k2**2

    if LifetimeModel == 2: # efficient approximation to the hypergeometric function
        kL[kL < 0.005] = 0.005
        kLsquare = kL**2
        LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
        ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
        ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
        Beta = gamma_par*LifeTime
    elif LifetimeModel == 1:
        Beta = gamma_par/ (kL**(2/3))
    elif LifetimeModel == 0:
    # Numerical evaluation of the hypergeometric function
    # NB!! Not implemented in the Python version yet
        Beta = gamma_par/( (kL**(2/3)) * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

    k30 = k3 + Beta*k1
    k30square = k30**2
    k0square = k1square + k2square + k30square
    k0 = np.sqrt(k0square)
    kL0 = k0*L
    
    with np.errstate(divide='ignore', invalid='ignore'):
        C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
        
        C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
            np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
        
        Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
        
        zeta1 = np.asarray(C1 - (k2/k1)*C2)
        zeta2 = np.asarray((k2/k1)*C1 + C2)
        
        zeta1[k1 == 0] = -Beta[k1 == 0]
        zeta2[k1 == 0] = 0
    
    if ElementChoice == None:
        
        Phi11 = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k1square - 2*k1*k30*zeta1 + (k1square + k2square)*(zeta1**2)))
        Phi22 = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k2square - 2*k2*k30*zeta2 + (k1square + k2square)*(zeta2**2)))
        Phi33 = np.asarray((Ek0/(4*np.pi*(k**4)))*(k1square + k2square))
    
        Phi12 = np.asarray((Ek0/(4*np.pi*(k0**4)))*( -k1*k2 - k1*k30*zeta2 - k2*k30*zeta1 + (k1square + k2square)*zeta1*zeta2))
        Phi13 = np.asarray((Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k1*k30 + (k1square + k2square)*zeta1))
        Phi23 = np.asarray((Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k2*k30 + (k1square + k2square)*zeta2))
        
        Phi11[ksquare==0] = 0
        Phi22[ksquare==0] = 0
        Phi33[ksquare==0] = 0
        Phi12[ksquare==0] = 0
        Phi13[ksquare==0] = 0
        Phi23[ksquare==0] = 0
        
        if np.asarray(k1).size == 1:
            Phi = np.array([[Phi11, Phi12, Phi13],[Phi12, Phi22, Phi23],[Phi13,Phi23,Phi33]])
        else:
            Phi = np.array([Phi11,Phi22,Phi33,Phi12,Phi13,Phi23])
            
    else:
        
        if ElementChoice == 11:
            Phi = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k1square - 2*k1*k30*zeta1 + (k1square + k2square)*(zeta1**2)))
            Phi[ksquare==0] = 0
        elif ElementChoice == 22:
            Phi = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k2square - 2*k2*k30*zeta2 + (k1square + k2square)*(zeta2**2)))
            Phi[ksquare==0] = 0
        elif ElementChoice == 33:
            Phi = np.asarray((Ek0/(4*np.pi*(k**4)))*(k1square + k2square))
            Phi[ksquare==0] = 0
        elif ElementChoice == 12:
            Phi = np.asarray((Ek0/(4*np.pi*(k0**4)))*( -k1*k2 - k1*k30*zeta2 - k2*k30*zeta1 + (k1square + k2square)*zeta1*zeta2))
            Phi[ksquare==0] = 0
        elif ElementChoice == 13:
            Phi = np.asarray((Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k1*k30 + (k1square + k2square)*zeta1))
            Phi[ksquare==0] = 0
        elif ElementChoice == 23:
            Phi = np.asarray((Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k2*k30 + (k1square + k2square)*zeta2))
            Phi[ksquare==0] = 0
        
    return Phi

def manntensorcomponents(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel,ElementChoice = None):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    kL = np.asarray(k*L)
    k1square = k1**2
    k2square = k2**2

    if LifetimeModel == 2: # efficient approximation to the hypergeometric function
        kL[kL < 0.005] = 0.005
        kLsquare = kL**2
        LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
        ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
        ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
        Beta = gamma_par*LifeTime
    elif LifetimeModel == 1:
        Beta = gamma_par/ (kL**(2/3))
    elif LifetimeModel == 0:
        # Numerical evaluation of the hypergeometric function
        # NB!! Not implemented in the Python version yet
        Beta = gamma_par/( (kL**(2/3))  * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

    k30 = k3 + Beta*k1
    k30square = k30**2
    k0square = k1square + k2square + k30square
    k0 = np.sqrt(k0square)
    kL0 = k0*L
    
    with np.errstate(divide='ignore', invalid='ignore'):
        C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
        
        C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
            np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
    
        Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
        
        zeta1 = np.asarray(C1 - (k2/k1)*C2)
        zeta2 = np.asarray((k2/k1)*C1 + C2)
    
        zeta1[k1 == 0] = -Beta[k1 == 0]
        zeta2[k1 == 0] = 0
            
        Phi11 = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k1square - 2*k1*k30*zeta1 + (k1square + k2square)*(zeta1**2)))
        Phi22 = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k2square - 2*k2*k30*zeta2 + (k1square + k2square)*(zeta2**2)))
        Phi33 = np.asarray((Ek0/(4*np.pi*(k**4)))*(k1square + k2square))
    
        Phi12 = np.asarray((Ek0/(4*np.pi*(k0**4)))*( -k1*k2 - k1*k30*zeta2 - k2*k30*zeta1 + (k1square + k2square)*zeta1*zeta2))
        Phi13 = np.asarray((Ek0/(4*np.pi*k0square*ksquare))*( -k1*k30 + (k1square + k2square)*zeta1))
        Phi23 = np.asarray((Ek0/(4*np.pi*k0square*ksquare))*( -k2*k30 + (k1square + k2square)*zeta2))
        
        Phi11[ksquare==0] = 0
        Phi22[ksquare==0] = 0
        Phi33[ksquare==0] = 0
        Phi12[ksquare==0] = 0
        Phi13[ksquare==0] = 0
        Phi23[ksquare==0] = 0
    
    return Phi11,Phi22,Phi33,Phi12,Phi13,Phi23

def manntensor1component(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel,ElementChoice = None):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    kL = np.asarray(k*L)
    k1square = k1**2
    k2square = k2**2

    if LifetimeModel == 2: # efficient approximation to the hypergeometric function
        kL[kL < 0.005] = 0.005
        kLsquare = kL**2
        LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
        ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
        ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
        Beta = gamma_par*LifeTime
    elif LifetimeModel == 1:
        Beta = gamma_par/ (kL**(2/3))
    elif LifetimeModel == 0:
        # Numerical evaluation of the hypergeometric function
        # NB!! Not implemented in the Python version yet
        Beta = gamma_par/( (kL**(2/3))  * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

    k30 = k3 + Beta*k1
    k30square = k30**2
    k0square = k1square + k2square + k30square
    k0 = np.sqrt(k0square)
    kL0 = k0*L
    
    with np.errstate(divide='ignore', invalid='ignore'):
        Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
        if ElementChoice == 11:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )           
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta1 = np.asarray(C1 - (k2/k1)*C2)
            zeta1[k1 == 0] = -Beta[k1 == 0]
            Phi = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k1square - 2*k1*k30*zeta1 + (k1square + k2square)*(zeta1**2)))
        elif ElementChoice == 22:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )            
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta2 = np.asarray((k2/k1)*C1 + C2)
            zeta2[k1 == 0] = 0
            Phi = np.asarray((Ek0/(4*np.pi*(k0**4)))*( k0square - k2square - 2*k2*k30*zeta2 + (k1square + k2square)*(zeta2**2)))
        elif ElementChoice == 33:
            Phi = np.asarray((Ek0/(4*np.pi*(k**4)))*(k1square + k2square))
        elif ElementChoice == 12:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )            
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta1 = np.asarray(C1 - (k2/k1)*C2)
            zeta2 = np.asarray((k2/k1)*C1 + C2)
            zeta1[k1 == 0] = -Beta[k1 == 0]
            zeta2[k1 == 0] = 0
            Phi = np.asarray((Ek0/(4*np.pi*(k0**4)))*( -k1*k2 - k1*k30*zeta2 - k2*k30*zeta1 + (k1square + k2square)*zeta1*zeta2))
        elif ElementChoice == 13:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
            
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta1 = np.asarray(C1 - (k2/k1)*C2)
            zeta1[k1 == 0] = -Beta[k1 == 0]
            Phi = np.asarray((Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k1*k30 + (k1square + k2square)*zeta1))
        elif ElementChoice == 23:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
            
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta2 = np.asarray((k2/k1)*C1 + C2)
            zeta2[k1 == 0] = 0
            Phi = np.asarray((Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k2*k30 + (k1square + k2square)*zeta2))
            
    Phi[ksquare==0] = 0
    return Phi

def manntensor_scalark(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    
    if k==0:
        Phi = np.zeros(3,3)
    else:
        kL = k*L
        k1square = k1**2
        k2square = k2**2

        if LifetimeModel == 2: # efficient approximation to the hypergeometric function
            if kL < 0.005:
                kL = 0.005
            kLsquare = kL**2
            LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
            ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
            ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
            Beta = gamma_par*LifeTime
        elif LifetimeModel == 1:
            Beta = gamma_par/ (kL**(2/3))
        elif LifetimeModel == 0:
        # Numerical evaluation of the hypergeometric function
        # NB!! Not implemented in the Python version yet
            Beta = gamma_par/( (kL**(2/3)) * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

        k30 = k3 + Beta*k1
        k30square = k30**2
        k0square = k1square + k2square + k30square
        k0 = np.sqrt(k0square)
        kL0 = k0*L
        
        C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
        
        C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
            np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
        
        Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
        if k1 == 0:
            zeta1 = -Beta
            zeta2 = 0
        else:
            zeta1 = C1 - (k2/k1)*C2
            zeta2 = (k2/k1)*C1 + C2
        
        Phi11 = (Ek0/(4*np.pi*(k0**4)))*( k0square - k1square - 2*k1*k30*zeta1 + (k1square + k2square)*(zeta1**2))
        Phi22 = (Ek0/(4*np.pi*(k0**4)))*( k0square - k2square - 2*k2*k30*zeta2 + (k1square + k2square)*(zeta2**2))
        Phi33 = (Ek0/(4*np.pi*(k**4)))*(k1square + k2square)
    
        Phi12 = (Ek0/(4*np.pi*(k0**4)))*( -k1*k2 - k1*k30*zeta2 - k2*k30*zeta1 + (k1square + k2square)*zeta1*zeta2)
        Phi13 = (Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k1*k30 + (k1square + k2square)*zeta1)
        Phi23 = (Ek0/(4*np.pi*(k0square)*(ksquare)))*( -k2*k30 + (k1square + k2square)*zeta2)
        
        Phi = np.array([[Phi11, Phi12, Phi13],[Phi12, Phi22, Phi23],[Phi13,Phi23,Phi33]])
        
    return Phi

def manntensorsqrt(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel,ElementChoice = None):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    kL = np.asarray(k*L)
    k1square = k1**2
    k2square = k2**2

    if LifetimeModel == 2: # efficient approximation to the hypergeometric function
        kL[kL < 0.005] = 0.005
        kLsquare = kL**2
        LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
        ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
        ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
        Beta = gamma_par*LifeTime
    elif LifetimeModel == 1:
        Beta = gamma_par/ (kL**(2/3))
    elif LifetimeModel == 0:
    # Numerical evaluation of the hypergeometric function
    # NB!! Not implemented in the Python version yet
        Beta = gamma_par/( (kL**(2/3)) * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

    k30 = k3 + Beta*k1
    k30square = k30**2
    k0square = k1square + k2square + k30square
    k0 = np.sqrt(k0square)
    kL0 = k0*L
    
    C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
    
    C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
        np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
    
    Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
    
    zeta1 = np.asarray(C1 - (k2/k1)*C2)
    zeta2 = np.asarray((k2/k1)*C1 + C2)
    
    zeta1[k1 == 0] = -Beta[k1 == 0]
    zeta2[k1 == 0] = 0
    
    kL0 = k0*L
    Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
    Ek4 = np.sqrt(Ek0/(4*np.pi*(k0**4)))
    kk = (k0**2)/(k**2)
    
    Phi11 = Ek4*zeta1*k2
    Phi12 = Ek4*(k30 - zeta1*k1)
    Phi13 = Ek4*(-k2)
    Phi21 = Ek4*(-k30 + zeta2*k2)
    Phi22 = Ek4*(-zeta2*k1)
    Phi23 = Ek4*(k1)
    Phi31 = Ek4*(k2*kk)
    Phi32 = Ek4*(-k1*kk)
    Phi33 = k1*0
    if np.asarray(k1).size == 1:
        if k == 0:
            Phi = np.zeros(3,3)
        else:
            Phi = np.array([[Phi11,Phi12,Phi13],[Phi21,Phi22,Phi23],[Phi31,Phi32,Phi33]])
    else:
        Phi = np.array([Phi11,Phi12,Phi13,Phi21,Phi22,Phi23,Phi31,Phi32,Phi33])
    
    return Phi

def manntensorsqrtcomponents(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel,ElementChoice = None,VarianceRatios = None):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    kL = np.asarray(k*L)
    k1square = k1**2
    k2square = k2**2

    if LifetimeModel == 2: # efficient approximation to the hypergeometric function
        kL[kL < 0.005] = 0.005
        kLsquare = kL**2
        LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
        ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
        ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
        Beta = gamma_par*LifeTime
    elif LifetimeModel == 1:
        Beta = gamma_par/ (kL**(2/3))
    elif LifetimeModel == 0:
    # Numerical evaluation of the hypergeometric function
    # NB!! Not implemented in the Python version yet
        Beta = gamma_par/( (kL**(2/3)) * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

    k30 = k3 + Beta*k1
    k30square = k30**2
    k0square = k1square + k2square + k30square
    k0 = np.sqrt(k0square)
    
    with np.errstate(divide='ignore', invalid='ignore'):

    
        C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
        
        C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
            np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
                
        zeta1 = np.asarray(C1 - (k2/k1)*C2)
        zeta2 = np.asarray((k2/k1)*C1 + C2)
        
        zeta1[k1 == 0] = -Beta[k1 == 0]
        zeta2[k1 == 0] = 0
        
        kL0 = k0*L
        Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
        Ek4 = np.sqrt(Ek0/(4*np.pi*(k0**4)))
        kk = (k0**2)/(k**2)
        
    if VarianceRatios is None:
        Phi11 = Ek4*zeta1*k2
        Phi12 = Ek4*(k30 - zeta1*k1)
        Phi13 = Ek4*(-k2)
        Phi21 = Ek4*(-k30 + zeta2*k2)
        Phi22 = Ek4*(-zeta2*k1)
        Phi23 = Ek4*(k1)
        Phi31 = Ek4*(k2*kk)
        Phi32 = Ek4*(-k1*kk)
        Phi33 = k1*0        
    else:
        Phi11 = Ek4*zeta1*k2
        Phi12 = Ek4*(k30*VarianceRatios[0] - zeta1*k1)
        Phi13 = Ek4*(-k2*VarianceRatios[0])
        Phi21 = Ek4*(-k30*VarianceRatios[1] + zeta2*k2)
        Phi22 = Ek4*(-zeta2*k1)
        Phi23 = Ek4*(k1*VarianceRatios[1])
        Phi31 = Ek4*(k2*kk*VarianceRatios[2])
        Phi32 = Ek4*(-k1*kk*VarianceRatios[2])
        Phi33 = k1*0               

        
    
    Phi11[ksquare==0] = 0
    Phi12[ksquare==0] = 0
    Phi13[ksquare==0] = 0
    Phi21[ksquare==0] = 0
    Phi22[ksquare==0] = 0    
    Phi23[ksquare==0] = 0
    Phi31[ksquare==0] = 0
    Phi32[ksquare==0] = 0
    Phi33[ksquare==0] = 0
    
    return Phi11,Phi12,Phi13,Phi21,Phi22,Phi23,Phi31,Phi32,Phi33

def manntensorsqrt_scalark(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel,Phi = None):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    if Phi is None:
        Phi = np.zeros(3,3)
    
    if k==0:
        Phi = np.zeros(3,3)
    else:
        kL = k*L
        k1square = k1**2
        k2square = k2**2

        if LifetimeModel == 2: # efficient approximation to the hypergeometric function
            if kL < 0.005:
                kL = 0.005
            kLsquare = kL**2
            LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
            ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
            ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
            Beta = gamma_par*LifeTime
        elif LifetimeModel == 1:
            Beta = gamma_par/ (kL**(2/3))
        elif LifetimeModel == 0:
        # Numerical evaluation of the hypergeometric function
        # NB!! Not implemented in the Python version yet
            Beta = gamma_par/( (kL**(2/3)) * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

        k30 = k3 + Beta*k1
        k30square = k30**2
        k0square = k1square + k2square + k30square
        k0 = np.sqrt(k0square)
        kL0 = k0*L
        with np.errstate(divide='ignore', invalid='ignore'):
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )
            
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            
            Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
            if k1 == 0:
                zeta1 = -Beta
                zeta2 = 0
            else:
                zeta1 = C1 - (k2/k1)*C2
                zeta2 = (k2/k1)*C1 + C2
    
        kL0 = k0*L
        Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
        Ek4 = np.sqrt(Ek0/(4*np.pi*(k0**4)))
        kk = (k0**2)/(k**2)
        
        Phi11 = Ek4*zeta1*k2
        Phi12 = Ek4*(k30 - zeta1*k1)
        Phi13 = Ek4*(-k2)
        Phi21 = Ek4*(-k30 + zeta2*k2)
        Phi22 = Ek4*(-zeta2*k1)
        Phi23 = Ek4*(k1)
        Phi31 = Ek4*(k2*kk)
        Phi32 = Ek4*(-k1*kk)
        Phi33 = k1*0
        
        Phi[0,0] = Phi11
        Phi[0,1] = Phi12
        Phi[0,2] = Phi13
        Phi[1,0] = Phi21
        Phi[1,1] = Phi22
        Phi[1,2] = Phi23
        Phi[2,0] = Phi31
        Phi[2,1] = Phi32
        Phi[2,2] = Phi33

        #Phi = np.array([[Phi11,Phi12,Phi13],[Phi21,Phi22,Phi23],[Phi31,Phi32,Phi33]])
        
    return Phi

def manntensorsqrt1component(k1,k2,k3,gamma_par,L,alphaepsilon,LifetimeModel,ElementChoice = None):
    ksquare = k1**2 + k2**2 + k3**2
    k = np.sqrt( ksquare )
    kL = np.asarray(k*L)
    k1square = k1**2
    k2square = k2**2

    if LifetimeModel == 2: # efficient approximation to the hypergeometric function
        kL[kL < 0.005] = 0.005
        kLsquare = kL**2
        LifeTime = (((1 + kLsquare)**(1/6))/kL)* \
        ( ( 1.2050983316598936 - 0.04079766636961979*kL + 1.1050803451576134*kLsquare)/ \
        ( 1 - 0.04103886513006046*kL + 1.1050902034670118*kLsquare))
        Beta = gamma_par*LifeTime
    elif LifetimeModel == 1:
        Beta = gamma_par/ (kL**(2/3))
    elif LifetimeModel == 0:
    # Numerical evaluation of the hypergeometric function
    # NB!! Not implemented in the Python version yet
        Beta = gamma_par/( (kL**(2/3)) * np.sqrt( HypergeometricFunction(1/3,17/6,4/3,-(kL**(-2)),1,50) ) )

    k30 = k3 + Beta*k1
    k30square = k30**2
    k0square = k1square + k2square + k30square
    k0 = np.sqrt(k0square)
    kL0 = k0*L
    
    with np.errstate(divide='ignore', invalid='ignore'):
        Ek0 = alphaepsilon*(L**(5/3))*(kL0**4)/((1 + kL0**2)**(17/6))
        Ek4 = np.sqrt(Ek0/(4*np.pi*(k0**4)))

        if ElementChoice == 11:
    
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )            
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )            
            zeta1 = np.asarray(C1 - (k2/k1)*C2)            
            zeta1[k1 == 0] = -Beta[k1 == 0]            
            Phi = Ek4*zeta1*k2                                    
        elif ElementChoice == 12:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )            
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta1 = np.asarray(C1 - (k2/k1)*C2)
            zeta1[k1 == 0] = -Beta[k1 == 0]
            Phi = Ek4*(k30 - zeta1*k1)            
        elif ElementChoice == 13:
            Phi = Ek4*(-k2)
        elif ElementChoice == 21:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )           
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta2 = np.asarray((k2/k1)*C1 + C2)
            zeta2[k1 == 0] = 0            
            Phi = Ek4*(-k30 + zeta2*k2)
        elif ElementChoice == 22:
            C1 = Beta * k1square * (k0square - 2*k30square + Beta * k1 * k30) / ( (ksquare)*(k1square + k2square) )           
            C2 = (k2*(k0square)*( (k1square + k2square)**(-3/2) ))* \
                np.arctan2( Beta*k1*np.sqrt( k1square + k2square),( (k0square) - k30*k1*Beta ) )
            zeta2 = np.asarray((k2/k1)*C1 + C2)
            zeta2[k1 == 0] = 0
            Phi = Ek4*(-zeta2*k1)            
        elif ElementChoice == 23:            
            Phi = Ek4*(k1)
        elif ElementChoice == 31:
            kk = (k0**2)/(k**2)
            Phi = Ek4*(k2*kk)
        elif ElementChoice == 32:
            kk = (k0**2)/(k**2)
            Phi = Ek4*(-k1*kk)
        elif ElementChoice == 33:
            Phi = k1*0
    Phi[ksquare==0] = 0
    return Phi

def HypergeometricFunction(a,b,c,z):
    # Not implemented yet
    return 0



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 22 10:37:43 2019

@author: nkdi
"""
import numpy as np
def relu(x): return x*(x > 0)

def softplus(x): return np.log(1 + np.exp(x))

def logistic(x): return 1/(1 + np.exp(-x))

def hyperbolic_tangent(x): return (np.exp(2*x) - 1)/(np.exp(2*x) + 1)

def pairwise_correlation(A,B):
    N = B.shape[0]
    M = B.shape[1]
    muA = A.sum(0)/N
    muB = B.sum(0)/N
    muAB = (A*B).sum(0)/N
    muA2 = (A**2).sum(0)/N
    muB2 = (B**2).sum(0)/N
    rho = np.array([None]*M,ndmin = 2,dtype = 'float64')
    rho[0,:M] = (muAB - muA*muB)/(np.sqrt( muA2 - muA*muA)*np.sqrt( muB2 - muB*muB))
    return rho

def NormalDist(task,x,mu=0,sigma=1):
    import numpy as np
    if task == 0: # PDF
        y = (1.0/(sigma*np.sqrt(2.0*np.pi)))*np.exp(-((x - mu)**2)/(2.0*(sigma**2)))
    elif task == 1: # Cumulative
        from scipy.special import erf
        y = 0.5*(1.0 + erf((x - mu)/(sigma*np.sqrt(2))))
    elif task == 2: # Inverse
        from scipy.special import erfinv
        y = mu + sigma*np.sqrt(2)*erfinv(2*x - 1)        
    return y

def LogNormDist(task,x,mu,sigma):
    import numpy as np
    Eps   = np.sqrt(np.log( 1.0+(sigma/mu)**2 ) )
    Ksi   = np.log(mu)-0.5*Eps**2
    if task == 0: # PDF
        x[x<=0] = 1e-8
        u =(np.log(x)-Ksi)/Eps
        y = np.exp(-u*u/2.0)/(Eps*x*np.sqrt(2.0*np.pi))
    elif task == 1: # Cummulative
        x[x<=0] = 1e-8
        u =(np.log(x)-Ksi)/Eps
        y= NormalDist(1, u)
    elif task == 2: # Inverse
        y= np.exp(Ksi+Eps*NormalDist(2, x))
    
    return y

def CircularMean_D(x):
    x = np.asarray(x[~np.isnan(np.asarray(x,dtype = 'float64'))],dtype = 'float64')
    mu = np.arctan2(np.mean(np.sin(x*np.pi/180)),np.mean(np.cos(x*np.pi/180)))*180/np.pi
    mu+=360*((mu<=-0.5)==True)
    return mu



# -*- coding: utf-8 -*-
"""
Created on Mon May  3 11:39:55 2021

@author: nkdi
"""

def constrain_field(Constraints,u = None,v = None,w = None, BaseName = 'Turb', \
                        alphaepsilon = 1,L = 30,Gamma = 3, Nx = 8192, Ny = 32, Nz = 32, \
                        dx = 1,dy = 1,dz = 1,SaveToFile = 0, UseNormalization = 0, TurbOptions = None):

    import numpy as np
    import time

    '''
    Parse inputs 
    '''
    
    if TurbOptions is None:
        SeedNo = 0
        HighFreqComp = 0
        FileFormat = 0
    else:
        if 'SeedNo' not in TurbOptions:
            SeedNo = 0
        else:
            SeedNo = TurbOptions['SeedNo']
            
        if 'HighFreqComp' not in TurbOptions:
            HighFreqComp = 0
        else:
            HighFreqComp = TurbOptions['HighFreqComp']
            
        if 'FileFormat' not in TurbOptions:
            FileFormat = 0
        else:
            FileFormat = TurbOptions['FileFormat']
    
    if ((u is None) | (v is None) | (w is None)):
        if 'Udatafile' not in TurbOptions:
            print('Source turbulence boxes for constrained simulation do not exist.')
            print('Generating turbulence boxes:')
            '''
            ======================================================================================
            If not given as inputs, load turbulence boxes from files or generate them if necessary
            ======================================================================================'''
            from hipersim.turbgen.generate_field import generate_field 
            u,v,w = generate_field(BaseName,alphaepsilon,L,Gamma,SeedNo,Nx,Ny,Nz,dx,dy,dz, \
                                    HighFreqComp, SaveToFile = 0)
        else:
            '''
            Load source turbulence boxes 
            '''
            u = np.fromfile(TurbOptions['Udatafile'],dtype = 'single').reshape((Nx,Ny,Nz))
            v = np.fromfile(TurbOptions['Udatafile'],dtype = 'single').reshape((Nx,Ny,Nz))
            w = np.fromfile(TurbOptions['Udatafile'],dtype = 'single').reshape((Nx,Ny,Nz))
            

    # from generate_field import generate_field
    # Nx = 1024
    # Ny = 32
    # Nz = 32
    # dx = 2
    # dy = 4
    # dz = 4
    # L = 30
    # Gamma = 3
    # alphaepsilon = 1
    # SeedNo = 10
    # HighFreqComp = 1
    # SaveToFile = 1
    # BaseName = 'Turb'
    # HighFreqCompCheat = 1
    # UseNormalization = 0
    # u,v,w = generate_field(BaseName,alphaepsilon,L,Gamma,SeedNo,Nx,Ny,Nz,dx,dy,dz, \
    #                         HighFreqComp, SaveToFile = 0)
    # Xconstraints = np.atleast_2d(np.arange(1,2000,4)).T
    # Yconstraints1 = 60*np.ones(Xconstraints.shape)
    # Zconstraints1 = 10*np.ones(Xconstraints.shape)
    # Yconstraints2 = 10*np.ones(Xconstraints.shape)
    # Zconstraints2 = 60*np.ones(Xconstraints.shape)
    # Uconstraints = 10*np.exp(-0.5*( (Xconstraints - 1000)/200)**2)
    # Vconstraints = -5*np.exp(-0.5*( (Xconstraints - 1000)/200)**2)
    # Wconstraints = -2*np.exp(-0.5*( (Xconstraints - 1000)/200)**2)
    # Constraints = np.concatenate([np.concatenate([Xconstraints, Xconstraints]), 
    #                               np.concatenate([Yconstraints1, Yconstraints2]),
    #                               np.concatenate([Zconstraints1, Zconstraints2]),
    #                               np.concatenate([Uconstraints, -Uconstraints]),
    #                               np.concatenate([Vconstraints, -Vconstraints]),
    #                               np.concatenate([Wconstraints, -Wconstraints])],axis = 1)
        

    
    ''' 
    ===============================================
    APPLY CONSTRAINTS
    ==============================================='''
    
    '''
    ====================================================================
     COMPUTE MANN TENSOR VALUES, THEN IFFT TO GET CORRELATIONS
    ===================================================================='''
    print('Computing Mann tensor / correlation arrays for constrained simulation:')
    
    from hipersim.turbgen.manntensor import manntensorcomponents
    tstart = time.time()
    R0uu = np.zeros((Nx,2*Ny,2*Nz), dtype = 'cfloat')
    R0vv = np.zeros((Nx,2*Ny,2*Nz), dtype = 'cfloat')
    R0ww = np.zeros((Nx,2*Ny,2*Nz), dtype = 'cfloat')
    R0uw = np.zeros((Nx,2*Ny,2*Nz), dtype = 'cfloat')
    
    '''
    The u-v, and v-w components are not considered because Rho_uv and Rho_vw
    are zero in the Mann turbulence model
    '''
    pi = np.pi
    k1sim = np.concatenate([np.arange(0,Nx),np.arange(-Nx,0)])*(pi/(Nx*dx))
    k2sim = np.concatenate([np.arange(0,Ny),np.arange(-Ny,0)])*(pi/(Ny*dy))
    k3sim = np.concatenate([np.arange(0,Nz),np.arange(-Nz,0)])*(pi/(Nz*dz))
    
    k2simgrid,k3simgrid = np.meshgrid(k2sim,k3sim)
    '''
    Only half of the wave numbers are considered, it is considered that
    the correlation will be symmetric about k1 = 0 
    '''
    for ik1 in range(Nx):    
        Phi11ij, Phi22ij, Phi33ij, __, Phi13ij, __ = manntensorcomponents(k1sim[ik1]*np.ones(k2simgrid.shape),k2simgrid,k3simgrid,Gamma,L,alphaepsilon,2)
        R0uu[ik1,:,:] = np.fft.ifft2(Phi11ij,axes = (0,1))
        R0vv[ik1,:,:] = np.fft.ifft2(Phi22ij,axes = (0,1))
        R0ww[ik1,:,:] = np.fft.ifft2(Phi33ij,axes = (0,1))
        R0uw[ik1,:,:] = np.fft.ifft2(Phi13ij,axes = (0,1))
                   
    Ruu = np.zeros((2*Nx,Ny,Nz))
    Rvv = np.zeros((2*Nx,Ny,Nz))
    Rww = np.zeros((2*Nx,Ny,Nz))
    Ruw = np.zeros((2*Nx,Ny,Nz))
    for ik2 in range(Ny):
        for ik3 in range(Nz):
            ''' 
            !!! Needs to be checked !!!
            Changing the choice of where we apply the complex conjugate ends up with 
            somewhat different constrained fields '''
            Ruu[:,ik2,ik3] = np.real(np.fft.ifft(np.concatenate([np.conj(R0uu[:,ik2,ik3]), np.flip((R0uu[:,ik2,ik3]))])))
            Rvv[:,ik2,ik3] = np.real(np.fft.ifft(np.concatenate([np.conj(R0vv[:,ik2,ik3]), np.flip((R0vv[:,ik2,ik3]))])))
            Rww[:,ik2,ik3] = np.real(np.fft.ifft(np.concatenate([np.conj(R0ww[:,ik2,ik3]), np.flip((R0ww[:,ik2,ik3]))])))
            Ruw[:,ik2,ik3] = np.real(np.fft.ifft(np.concatenate([np.conj(R0uw[:,ik2,ik3]), np.flip((R0uw[:,ik2,ik3]))])))
    
    Ruu = Ruu[:Nx,:Ny,:Nz]
    Rvv = Rvv[:Nx,:Ny,:Nz]
    Rww = Rww[:Nx,:Ny,:Nz]
    Ruw = Ruw[:Nx,:Ny,:Nz]
    
    Ruw = Ruw/( np.sqrt(Ruu[0,0,0]*Rww[0,0,0]) )
    Ruu = Ruu/Ruu[0,0,0]
    Rvv = Rvv/Rvv[0,0,0]
    Rww = Rww/Rww[0,0,0]
    
    del R0uu, R0vv, R0ww, R0uw # Clear memory from unnecessary variables        
            
    t1 = time.time()
    print('Correlation computations complete')
    print('Time elapsed is ' + str(t1-tstart))
    
    '''
    ================================================================
     Compute distances, normalize wind fields and constraint fields
    ================================================================'''
    
    
    if UseNormalization == 1:
        varUdata = np.var(u)
        muUdata = np.mean(u)
        stdUdata = np.sqrt(varUdata)
        Unorm = (u - muUdata)/varUdata
        varVdata = np.var(v)
        muVdata = np.mean(v)
        stdVdata = np.sqrt(varVdata)
        Vnorm = (v - muVdata)/varVdata
        varWdata = np.var(w)
        muWdata = np.mean(w)
        stdWdata = np.sqrt(varWdata)
        Wnorm = (w - muWdata)/varWdata    
        ConstraintValuesUNorm = (Constraints[:,3] - muUdata)/stdUdata
        ConstraintValuesVNorm = (Constraints[:,4] - muVdata)/stdVdata
        ConstraintValuesWNorm = (Constraints[:,5] - muWdata)/stdWdata
    else:
        ConstraintValuesUNorm = Constraints[:,3]
        ConstraintValuesVNorm = Constraints[:,4]
        ConstraintValuesWNorm = Constraints[:,5]
        Unorm = u
        Vnorm = v
        Wnorm = w
    
    #del u, v, w
    
    '''
    ==========================================
     ASSEMBLE CONSTRAINT COVARIANCE MATRIX
    =========================================='''
    print('Populating covariance matrix for the constraints:')
    Clocx = np.rint(Constraints[:,0]/dx)
    Clocy = np.rint(Constraints[:,1]/dy)
    Clocz = np.rint(Constraints[:,2]/dz)
    '''
    ---------------------------------
     Eliminate overlapping constraints
    ---------------------------------'''
    ClocA = np.concatenate([np.atleast_2d(Clocx), np.atleast_2d(Clocy), np.atleast_2d(Clocz)]).T
    __,ClocIndex = np.unique(ClocA,axis=0,return_index = True)
    Constraints = Constraints[ClocIndex,:]
    Clocx = Clocx[ClocIndex]
    Clocy = Clocy[ClocIndex]
    Clocz = Clocz[ClocIndex]
    ConstraintValuesUNorm = ConstraintValuesUNorm[ClocIndex]
    ConstraintValuesVNorm = ConstraintValuesVNorm[ClocIndex]
    ConstraintValuesWNorm = ConstraintValuesWNorm[ClocIndex]
    Nconstraints = Constraints.shape[0]
    
    '''
    ---------------------------------------------
     Eliminate constraints too close to each other
    ---------------------------------------------'''
    Xdist = np.dot(np.atleast_2d(Constraints[:,0]).T,np.ones([1,Nconstraints])) - np.dot(np.ones([Nconstraints,1]),np.atleast_2d(Constraints[:,0]))
    Ydist = np.dot(np.atleast_2d(Constraints[:,1]).T,np.ones([1,Nconstraints])) - np.dot(np.ones([Nconstraints,1]),np.atleast_2d(Constraints[:,1]))
    Zdist = np.dot(np.atleast_2d(Constraints[:,2]).T,np.ones([1,Nconstraints])) - np.dot(np.ones([Nconstraints,1]),np.atleast_2d(Constraints[:,2]))
    
    Rdist = np.sqrt(Xdist**2 + Ydist**2 + Zdist**2)
    Rlimit = max( [L/10, min([dx, dy, dz])])
    Rexceed = (Rdist > 0) & (Rdist < Rlimit)
    ValidDistIndex = np.full((Rdist.shape[0]),True,dtype = 'bool')
    for i in range(Nconstraints):
        if np.any(Rexceed[i,:i]):
            Rexceed[i,:] = 0
            Rexceed[:,i] = 0
            ValidDistIndex[i] = False
    Constraints = Constraints[ValidDistIndex,:]
    Clocx = Clocx[ValidDistIndex].astype('int')
    Clocy = Clocy[ValidDistIndex].astype('int')
    Clocz = Clocz[ValidDistIndex].astype('int')
    ConstraintValuesUNorm = ConstraintValuesUNorm[ValidDistIndex]
    ConstraintValuesVNorm = ConstraintValuesVNorm[ValidDistIndex]
    ConstraintValuesWNorm = ConstraintValuesWNorm[ValidDistIndex]
    Nconstraints = Constraints.shape[0]
    
    del Rdist, Rexceed, ValidDistIndex
    
    
    '''
    -----------------------------
     Assemble u-v-w-correlation matrix
    -----------------------------'''
    CorrCMannUVW = np.zeros((3*Nconstraints,3*Nconstraints))
    
    for iC in range(Nconstraints):
        xloci = Clocx[iC]
        yloci = Clocy[iC]
        zloci = Clocz[iC]
        xlocCij = np.abs(xloci-Clocx).astype('int')
        ylocCij = np.abs(yloci-Clocy).astype('int')
        zlocCij = np.abs(zloci-Clocz).astype('int')
        
        CorrUUij = Ruu[xlocCij,ylocCij,zlocCij]
        CorrVVij = Rvv[xlocCij,ylocCij,zlocCij]
        CorrWWij = Rww[xlocCij,ylocCij,zlocCij]
        CorrUWij = Ruw[xlocCij,ylocCij,zlocCij]
    
        # CorrUUij = np.zeros(Nconstraints)
        # CorrVVij = np.zeros(Nconstraints)
        # CorrWWij = np.zeros(Nconstraints)
        # CorrUWij = np.zeros(Nconstraints)
        # for jC in range(Nconstraints):
        #     CorrUUij[jC] = Ruu[xlocCij[jC],ylocCij[jC],zlocCij[jC]]
        #     CorrVVij[jC] = Rvv[xlocCij[jC],ylocCij[jC],zlocCij[jC]]
        #     CorrWWij[jC] = Rww[xlocCij[jC],ylocCij[jC],zlocCij[jC]]
        #     CorrUWij[jC] = Ruw[xlocCij[jC],ylocCij[jC],zlocCij[jC]]
    
        CorrCMannUVW[:Nconstraints,iC] = CorrUUij
        CorrCMannUVW[iC,:Nconstraints] = CorrUUij
        CorrCMannUVW[Nconstraints:2*Nconstraints,Nconstraints + iC] = CorrVVij
        CorrCMannUVW[Nconstraints + iC,Nconstraints:2*Nconstraints] = CorrVVij    
        CorrCMannUVW[2*Nconstraints:,2*Nconstraints + iC] = CorrWWij
        CorrCMannUVW[2*Nconstraints + iC,2*Nconstraints:] = CorrWWij    
        CorrCMannUVW[2*Nconstraints:,iC] = CorrUWij
        CorrCMannUVW[iC,2*Nconstraints:] = CorrUWij    
           
        
    t2 = time.time()
    print('Constraint-constraint covariance matrix has been assembled')
    print('Time elapsed is ' + str(t2-t1))
    
    '''
    =========================================================
     APPLY CONSTRAINT EQUATIONS TO COMPUTE THE RESIDUAL FIELD
    =========================================================
     Using eq.(2) from Dimitrov & Natarajan (2016) to compute 
     the residual field which is to be added to the unconstrained
     field in order to obtain the constrained result.
    
     Due to memory limitations, eq. (2) is evaluated in batches intended
     to avoid the need of fully assembling the first term in the equation,
     which is a matrix with size (Nx*Ny*Nz)x(Nconstraints).
     First, the product of the second and third term is evaluated, 
     then it is multiplied piecewise to a subset of the rows of the 
     first term.
    '''
    print('Applying constraints...')
    ConstraintValuesUVWNorm = np.concatenate([ConstraintValuesUNorm, ConstraintValuesVNorm, ConstraintValuesWNorm])
    UVWcontemporaneous = np.zeros((3*Nconstraints))
    UVWcontemporaneous[:Nconstraints] = Unorm[Clocx,Clocy,Clocz]
    UVWcontemporaneous[Nconstraints:2*Nconstraints] = Vnorm[Clocx,Clocy,Clocz]
    UVWcontemporaneous[2*Nconstraints:] = Wnorm[Clocx,Clocy,Clocz]
    
    '''
     Computing the product of the second and third terms in eq.(2) in 
     Dimitrov & Natarajan (2016)
    '''
    
    CConstUVW = np.linalg.solve(CorrCMannUVW, (ConstraintValuesUVWNorm - UVWcontemporaneous))
    # CConstUVW = np.dot(np.linalg.inv(CorrCMannUVW), ConstraintValuesUVWNorm - UVWcontemporaneous)
    
    CConstU = np.atleast_2d(CConstUVW[:Nconstraints]).T
    CConstV = np.atleast_2d(CConstUVW[Nconstraints:2*Nconstraints]).T
    CConstW = np.atleast_2d(CConstUVW[2*Nconstraints:]).T
    del CorrCMannUVW
    
    Ruu = Ruu.reshape(Nx*Ny*Nz)
    Rvv = Rvv.reshape(Nx*Ny*Nz)
    Rww = Rww.reshape(Nx*Ny*Nz)
    Ruw = Ruw.reshape(Nx*Ny*Nz)
    
    ygrid, xgrid, zgrid = np.meshgrid(np.arange(Ny),np.arange(Nx),np.arange(Nz))
    xvect = xgrid.reshape(Nx*Ny*Nz)
    yvect = ygrid.reshape(Nx*Ny*Nz)
    zvect = zgrid.reshape(Nx*Ny*Nz)
    
    del xgrid, ygrid, zgrid
    
    ures = np.zeros(Nx*Ny*Nz)
    vres = np.zeros(Nx*Ny*Nz)
    wres = np.zeros(Nx*Ny*Nz)
    
    # Npoints = Nx*Ny*Nz
    
    for iC in range(Nconstraints):
        # dxCov = np.abs(xvect - Clocx[iC])
        # dyCov = np.abs(yvect - Clocy[iC])
        # dzCov = np.abs(zvect - Clocz[iC])
        dlinear = np.abs(xvect - Clocx[iC])*(Ny*Nz) + np.abs(yvect - Clocy[iC])*Ny + np.abs(zvect - Clocz[iC])
        CorrUUi = Ruu[dlinear]
        CorrVVi = Rvv[dlinear]
        CorrWWi = Rww[dlinear]
        CorrUWi = Ruw[dlinear]
        ures+= CorrUUi*CConstU[iC] + CorrUWi*CConstW[iC]
        vres+= CorrVVi*CConstV[iC]
        wres+= CorrUWi*CConstU[iC] + CorrWWi*CConstW[iC]
        
    Uconstrained = Unorm + np.reshape(ures,(Nx,Ny,Nz))
    Vconstrained = Vnorm + np.reshape(vres,(Nx,Ny,Nz))
    Wconstrained = Wnorm + np.reshape(wres,(Nx,Ny,Nz))
    
    
    '''
    OLDER IMPLEMENTATION - SIGNIFICANTLY FASTER IN MATLAB BUT SLOW IN PYTHON
    DUE TO ARRAY INDEXING SPEED '''
    
    # Uconstrained = np.zeros((Nx,Ny,Nz))
    # Vconstrained = np.zeros((Nx,Ny,Nz))
    # Wconstrained = np.zeros((Nx,Ny,Nz))   
    
    # CorrUUxyz = np.zeros((Nx,Nconstraints))
    # CorrVVxyz = np.zeros((Nx,Nconstraints))
    # CorrUVxyz = np.zeros((Nx,Nconstraints))
    # CorrVWxyz = np.zeros((Nx,Nconstraints))
    # CorrUWxyz = np.zeros((Nx,Nconstraints))
    # CorrWWxyz = np.zeros((Nx,Nconstraints)) 
    
    # for yloc in range(Ny):
    #     for zloc in range(Nz):
    #         xloc = np.arange(Nx)
    #         dyCov = np.abs(yloc - Clocy).astype('int')
    #         dzCov = np.abs(zloc - Clocz).astype('int')
    #         '''
    #         !!! This loop needs optimization !!!
    #         '''
    #         for iC in range(Nconstraints):
    #             dxCov = np.abs(xloc - Clocx[iC])
    #             CorrUUxyz[:,iC] = Ruu[dxCov,dyCov[iC],dzCov[iC]]
    #             CorrVVxyz[:,iC] = Rvv[dxCov,dyCov[iC],dzCov[iC]]
    #             CorrUWxyz[:,iC] = Ruw[dxCov,dyCov[iC],dzCov[iC]]
    #             CorrWWxyz[:,iC] = Rww[dxCov,dyCov[iC],dzCov[iC]]
    
    #         '''
    #           Computing the product of the first term in eq. (2) with the
    #           second and third terms - for one point in the y-z plane at a
    #           time.
    #         '''        
    #         CtermUYZ = np.dot(CorrUUxyz,CConstU) + np.dot(CorrUVxyz,CConstV) + np.dot(CorrUWxyz,CConstW) # Residual field term - u
    #         CtermVYZ = np.dot(CorrUVxyz,CConstU) + np.dot(CorrVVxyz,CConstV) + np.dot(CorrVWxyz,CConstW) # Residual field term - v
    #         CtermWYZ = np.dot(CorrUWxyz,CConstU) + np.dot(CorrVWxyz,CConstV) + np.dot(CorrWWxyz,CConstW) # Residual field term - w
    
    #         Uconstrained[:,yloc,zloc] = Unorm[:,yloc,zloc] + np.squeeze(CtermUYZ)
    #         Vconstrained[:,yloc,zloc] = Vnorm[:,yloc,zloc] + np.squeeze(CtermVYZ)
    #         Wconstrained[:,yloc,zloc] = Wnorm[:,yloc,zloc] + np.squeeze(CtermWYZ)
    
    # if Options.UseNormalization == 1
    #     Uconstrained = Uconstrained.*stdUdata + muUdata;
    #     Vconstrained = Vconstrained.*stdVdata + muVdata;
    #     Wconstrained = Wconstrained.*stdWdata + muWdata;
    # end
    
    t3 = time.time()
    print('Constraint application complete')
    print('Time elapsed is ' + str(t3 - t2))
    
    if SaveToFile == 1:
        if (FileFormat == 0) | (FileFormat == 'Hawc2') | (FileFormat == 'Mann'):
            Uconstrained.reshape(Nx*Ny*Nz).astype('single').tofile(BaseName + '_' + str(SeedNo) + '_c_u' + '.bin',sep = '')
            Vconstrained.reshape(Nx*Ny*Nz).astype('single').tofile(BaseName + '_' + str(SeedNo) + '_c_v' + '.bin',sep = '')
            Wconstrained.reshape(Nx*Ny*Nz).astype('single').tofile(BaseName + '_' + str(SeedNo) + '_c_w' + '.bin',sep = '')
        elif (FileFormat == 'Bladed') | (FileFormat == 'wnd') | (FileFormat == 'TurbSim'):
            from hipersim.turbgen.turb_utils import output_field
            params = {'dx':dx, 'dy':dy, 'dz':dz, 'Nx':Nx, 'Ny': Ny, 'Nz':Nz}
            output_field(Uconstrained,Vconstrained,Wconstrained,params,TurbOptions)    
    
    tend = time.time()
    print('Constrained simulation complete')
    print('Total time elapsed is ' + str(tend-tstart))

    return Uconstrained, Vconstrained, Wconstrained